@@ -69,6 +69,7 @@
 #include <mach/vm_map.h>
 #include <mach/vm_types.h>
 #include <mach/vm_prot.h>
+#include <mach/shared_memory_server.h>
 
 #include <device/device_types.h>
 #include <CoreFoundation/CoreFoundation.h>
@@ -82,7 +83,8 @@
 
 #include <libc.h>
 #include <termios.h>
-#include <bsd/curses.h>
+#include <curses.h>
+#include <sys/ioctl.h>
 
 /* Number of lines of header information on the standard screen */
 #define	HEADER_LINES	8
@@ -120,7 +122,6 @@ int werase(WINDOW *win);
 int total_threads;
 unsigned long long total_fw_private;
 
-struct termios tmode, omode;
 char   bytesread[128];
 
 host_cpu_load_info_data_t lastcounters, curcounters, startcounters;
@@ -228,10 +229,6 @@ struct object_info      *shared_hash_table[OBJECT_TABLE_SIZE];
 
 struct object_info *of_free_list = 0;
 
-#define FW_CODE_BEG_ADDR  0x70000000
-#define FW_DATA_BEG_ADDR  0x80000000
-#define FW_DATA_END_ADDR  0x90000000
-
 /*
  *	Translate thread state to a number in an ordered scale.
  *	When collapsing all the threads' states to one for the
@@ -445,27 +442,27 @@ pmem_doit(task_port_t task, int pid, int *shared, int *private, int *aliased, in
 		mach_msg_type_number_t  count;
 	        vm_size_t		size;
 
-		if ( !split && (address >= FW_CODE_BEG_ADDR && address < FW_DATA_END_ADDR)) {
-		        vm_region_basic_info_data_64_t    b_info;
-			  
-		        count = VM_REGION_BASIC_INFO_COUNT_64;
-			if (err = vm_region_64(task, &address, &size, VM_REGION_BASIC_INFO, (vm_region_info_t)&b_info,
-					    &count, &object_name))
-			        break;
-
-			if (b_info.reserved)
-			        split = 1;
-		}
 		count = VM_REGION_TOP_INFO_COUNT;
 
 		if (err = vm_region(task, &address, &size, VM_REGION_TOP_INFO, (vm_region_info_t)&info,
 				    &count, &object_name))
 		        break;
 
-		if (address >= FW_CODE_BEG_ADDR && address < FW_DATA_END_ADDR) {
+		if (address >= GLOBAL_SHARED_TEXT_SEGMENT && address < (GLOBAL_SHARED_DATA_SEGMENT + SHARED_DATA_REGION_SIZE)) {
 
 			*fw_private += info.private_pages_resident * vm_page_size;
 
+			if ( !split && info.share_mode == SM_EMPTY) {
+			        vm_region_basic_info_data_64_t    b_info;
+			  
+				count = VM_REGION_BASIC_INFO_COUNT_64;
+				if (err = vm_region_64(task, &address, &size, VM_REGION_BASIC_INFO, (vm_region_info_t)&b_info,
+					    &count, &object_name))
+				        break;
+
+				if (b_info.reserved)
+				        split = 1;
+			}
 		        if (info.share_mode != SM_PRIVATE) {
 			        address += size;
 			        continue;
@@ -529,13 +526,13 @@ pmem_doit(task_port_t task, int pid, int *shared, int *private, int *aliased, in
 		}
 	}
 	if (split)
-	        *vsize -= (FW_DATA_END_ADDR - FW_CODE_BEG_ADDR);
+	        *vsize -= (SHARED_TEXT_REGION_SIZE + SHARED_DATA_REGION_SIZE);
 }
 
 
 void
 pmem_fw_resident(unsigned int *num_fw, unsigned long long *vsize, unsigned int *code_size, unsigned int *data_size, unsigned int *linkedit_size)
-{       vm_address_t	address = FW_CODE_BEG_ADDR;
+{       vm_address_t	address = GLOBAL_SHARED_TEXT_SEGMENT;
 	kern_return_t	err = 0;
 	int             state = 0;
 
@@ -545,7 +542,7 @@ pmem_fw_resident(unsigned int *num_fw, unsigned long long *vsize, unsigned int *
         *data_size = 0;
         *linkedit_size = 0;
 
-	while (address < FW_DATA_END_ADDR) {
+	while (address < (GLOBAL_SHARED_DATA_SEGMENT + SHARED_DATA_REGION_SIZE)) {
 	        vm_region_submap_info_data_64_t   s_info;
 		mach_msg_type_number_t         count;
 	        vm_size_t	size;
@@ -557,9 +554,9 @@ pmem_fw_resident(unsigned int *num_fw, unsigned long long *vsize, unsigned int *
 		if (err = vm_region_recurse_64(mach_task_self(), &address, &size, &nesting_depth, (vm_region_info_t)&s_info, &count))
 		        break;
 
-		if (address >= FW_DATA_END_ADDR)
+		if (address >= (GLOBAL_SHARED_DATA_SEGMENT + SHARED_DATA_REGION_SIZE))
 		        break;
-		if (address < FW_DATA_BEG_ADDR) {
+		if (address < GLOBAL_SHARED_DATA_SEGMENT) {
 
 			if (s_info.share_mode == SM_SHARED || s_info.share_mode == SM_COW) {
 			        if (s_info.max_protection & VM_PROT_EXECUTE) {
@@ -830,20 +827,23 @@ void get_proc_info(kpb, pi)
 				(void) vm_deallocate(mach_task_self(), (vm_offset_t)thread_table,
 						     table_size * sizeof(*thread_table));
 
-				if (mach_port_names(task, &names, &ncnt,
-						    &types, &tcnt) == KERN_SUCCESS) {
-				        pi->num_ports = ncnt;
-					pi->orig_num_ports = ncnt;
-					(void) vm_deallocate(mach_task_self(),
-							     (vm_offset_t) names,
-							     ncnt * sizeof(*names));
-					(void) vm_deallocate(mach_task_self(),
-							     (vm_offset_t) types,
-							     tcnt * sizeof(*types));
-				} else {
-				        pi->num_ports = -1;
-				}
-			
+				if (!events_only) {
+				        if (mach_port_names(task, &names, &ncnt,
+							    &types, &tcnt) == KERN_SUCCESS) {
+					        pi->num_ports = ncnt;
+						pi->orig_num_ports = ncnt;
+						(void) vm_deallocate(mach_task_self(),
+								     (vm_offset_t) names,
+								     ncnt * sizeof(*names));
+						(void) vm_deallocate(mach_task_self(),
+								     (vm_offset_t) types,
+								     tcnt * sizeof(*types));
+					} else {
+					        pi->num_ports = -1;
+					}
+				} else
+				        pi->num_ports = 0;
+
 				if (events_only) {
 				        task_events_info_data_t	tei;
 
@@ -881,8 +881,6 @@ void leave()			/* exit under normal conditions -- INT handler */
 	        move(LINES - 1, 0);
 		refresh();
 		endwin();
-
-		tcsetattr(0, TCSANOW, &omode);
 	}
 	exit(0);
 }
@@ -890,10 +888,8 @@ void leave()			/* exit under normal conditions -- INT handler */
 void quit(status)		/* exit under duress */
 	int status;
 {
-        if (!oneshot) {
+        if (!oneshot)
 	        endwin();
-		tcsetattr(0, TCSANOW, &omode);
-	}
 	exit(status);
 }
 
@@ -1202,6 +1198,7 @@ main(argc, argv)
 	char	*argv[];
 {
 	char	*myname = "top";
+	int     ch;
 	int	delay = Default_DELAY;
 	kern_return_t	error;
 
@@ -1225,19 +1222,11 @@ main(argc, argv)
 	events_delta = 0;
 	events_accumulate = 0;
 
-	while (argc > 1 && argv[1][0] == '-') {
-		switch (argv[1][1]) {
+	while ((ch = getopt(argc, argv, "uwks:edal:")) != EOF) {
+	       switch(ch) {
 		case 's':
-		        if (argv[1][2])
-			        delay = atoi(&argv[1][2]);
-			else {
-				argc--;
-				argv++;
-
-				if (argc > 1)
-				        delay = atoi(&argv[1][0]);
-			}
-			break;
+			delay = atoi(optarg);
+		        break;
 		case 'u':
 			sort_by_usage = 1;
 			break;
@@ -1259,21 +1248,13 @@ main(argc, argv)
 		        events_accumulate = 1;
 			break;
 	        case 'l':
-		        if (argv[1][2])
-			        logcnt = atoi(&argv[1][2]);
-			else {
-				argc--;
-				argv++;
-
-				if (argc > 1)
-				        logcnt = atoi(&argv[1][0]);
-			}
+			logcnt = atoi(optarg);
 		        oneshot = 1;
 			LINES = 80;
 			COLS  = 132;
 			break;
 		default:
-			fprintf(stderr, "Usage: %s [-u] [-w] [-k] [-sn] [-e] [-d] [number]\n", myname);
+			fprintf(stderr, "Usage: %s [-u] [-w] [-k] [-sn] [-e | -d | -a] [-ln] [number]\n", myname);
 			fprintf(stderr, "  -u      enables sort by usage\n");
 			fprintf(stderr, "  -w      enables wide output of additional info\n");
 			fprintf(stderr, "  -k      generate vm info for kernel(proc 0)... expensive\n");
@@ -1286,9 +1267,11 @@ main(argc, argv)
 
 			exit(1);
 		}
-		argc--;
-		argv++;
 	}
+
+	argc -= optind;
+	//argv += optind;
+
 	if (events_only)
 	  {
 	    if ( wide_output || do_proc0_vm)
@@ -1303,8 +1286,8 @@ main(argc, argv)
 	host_port = get_host_port();
 
 	/* get count of top processes to display (if any) */
-	if (argc > 1) {
-		wanted_topn = topn = atoi(argv[1]);
+	if (argc) {
+		wanted_topn = topn = atoi(argv[optind]);
 	} else
 	        wanted_topn = -1;
 
@@ -1323,22 +1306,10 @@ main(argc, argv)
 	/* initializes curses and screen (last) */
 
 	if (!oneshot) {
-	        if (tcgetattr(0, &tmode) < 0) {
-		        printf("can't get terminal attributes\n");
-			exit(1);
-		}
-		omode = tmode;
-	
-		tmode.c_lflag &= ~ICANON;
-		tmode.c_cc[VMIN] = 0;
-		tmode.c_cc[VTIME] = (delay * 10);
-
-		if (tcsetattr(0, TCSANOW, &tmode) < 0) {
-		        printf("can't set terminal attributes\n");
-			exit(1);
-		}
 		initscr();
-		refresh();
+		cbreak();
+		timeout(delay * 1000);
+		noecho();
 		erase();
 		clear();
 		refresh();
@@ -1397,12 +1368,16 @@ main(argc, argv)
 	        int   n;
 
 	        if (newLINES) {
+		        newLINES = 0;
 
 		        if (!oneshot) {
-			        initscr();
-				erase();
-				clear();
-				refresh();
+				struct winsize size;
+
+				if (ioctl(1, TIOCGWINSZ, &size) != -1) {
+					resizeterm(size.ws_row, size.ws_col);
+					erase();
+					clear();
+				}
 			}
 		        n = LINES - Header_lines;
 
@@ -1418,19 +1393,14 @@ main(argc, argv)
 					        topn = n;
 				}
 			}
-		        newLINES = 0;
 		}
 		(void)screen_update();
 
 		if (!oneshot) {
+			int	c;
 
-		        if ((n = read(0, &bytesread, 128)) > 0) {
-			        int   i;
-
-				for (i = 0; i < n; i++)
-				        if (bytesread[i] == 'q')
-					        leave();
-			}
+			if ((c = getch()) != ERR && (char)c == 'q') 
+				leave();
 		} else
 		        sleep(delay);
 	}
@@ -1448,7 +1418,7 @@ void screen_update()
 	unsigned long long	total_virtual_size;
 	unsigned long long	total_private_size;
 	unsigned long long      total_shared_size;
-	unsigned int            total_memory_regions;
+	unsigned int            total_memory_regions = 0;
 	unsigned int            total_shared_objects;
 	unsigned int            total_fw_code_size;
 	unsigned int            total_fw_data_size;
@@ -1492,9 +1462,11 @@ void screen_update()
 	        mach_error("host_info", error);
 	        exit(EXIT_FAILURE);
 	}
-	getNETWORKcounters();
-	getDISKcounters();
-
+	
+	if (events_only) {
+	        getNETWORKcounters();
+		getDISKcounters();
+	}
 	/* count up process states and get pointers to interesting procs */
 
 	mpid = 0;
@@ -1746,7 +1718,7 @@ void screen_update()
 		        i_kbytes = i_net.kbytes;
 		        o_kbytes = o_net.kbytes;
 		}
-	        sprintf(tbuf, "en0:   %10d ipkts/%dK", i_io, i_kbytes);
+	        sprintf(tbuf, "Networks:%10d ipkts/%dK", i_io, i_kbytes);
 		clen = strlen(tbuf);
 		memset(&tbuf[clen], ' ', 36 - clen);
 		sprintf(&tbuf[36], "%10d opkts /%dK\n", o_io, o_kbytes);
@@ -1792,7 +1764,7 @@ void screen_update()
 		        i_kbytes = i_dsk.kbytes;
 		        o_kbytes = o_dsk.kbytes;
 		}
-	        sprintf(tbuf, "Disks: %10d reads/%dK", i_io, i_kbytes);
+	        sprintf(tbuf, "Disks:   %10d reads/%dK", i_io, i_kbytes);
 		clen = strlen(tbuf);
 		memset(&tbuf[clen], ' ', 36 - clen);
 		sprintf(&tbuf[36], "%10d writes/%dK\n", o_io, o_kbytes);
@@ -1835,7 +1807,7 @@ void screen_update()
 		        pageins = vm_stat.pageins;
 		        pageouts = vm_stat.pageouts;
 		}
-	        sprintf(tbuf, "VM:    %10d pageins", pageins);
+	        sprintf(tbuf, "VM:      %10d pageins", pageins);
 		clen = strlen(tbuf);
 		memset(&tbuf[clen], ' ', 36 - clen);
 		sprintf(&tbuf[36], "%10d pageouts\n", pageouts);
@@ -2199,25 +2171,26 @@ getNETWORKcounters()
 	if (kread(nl_net[N_IFNET].n_value, (char *)&ifnethead, sizeof ifnethead))
 	       return;
 
+        i_net.io = 0;
+	o_net.io = 0;
+			  
+	i_net.kbytes = 0;
+	o_net.kbytes = 0;
+
 	for (off = (u_long)ifnethead.tqh_first; off; ) {
-                  char name[16], tname[16];
+                  char tname[16];
 
 		  if (kread(off, (char *)&ifnet, sizeof ifnet))
 		          break;
 		  if (kread((u_long)ifnet.if_name, tname, 16))
 		          break;
-                  tname[15] = '\0';
 
-		  snprintf(name, 16, "%s%d", tname, ifnet.if_unit);
-
-		  if (strcmp(name, "en0") == 0) {
-		          i_net.io = ifnet.if_ipackets;
-			  o_net.io = ifnet.if_opackets;
+		  if (strncmp(tname, "lo", 2)) {
+		          i_net.io += ifnet.if_ipackets;
+			  o_net.io += ifnet.if_opackets;
 			  
-			  i_net.kbytes = ifnet.if_ibytes/1024;
-			  o_net.kbytes = ifnet.if_obytes/1024;
-
-			  return;
+			  i_net.kbytes += ifnet.if_ibytes/1024;
+			  o_net.kbytes += ifnet.if_obytes/1024;
 		  }
 		  off = (u_long) ifnet.if_link.tqe_next;
 	}
@@ -2248,11 +2221,12 @@ getDISKcounters()
                                                             (CFMutableDictionaryRef *) &properties,
 	                                                    kCFAllocatorDefault,
 	                                                    kNilOptions);
+                if (properties) {
 
-		/* Obtain the statistics from the drive properties */
-	        statistics = (CFDictionaryRef) CFDictionaryGetValue(properties, CFSTR(kIOBlockStorageDriverStatisticsKey));
+                    /* Obtain the statistics from the drive properties */
+                    statistics = (CFDictionaryRef) CFDictionaryGetValue(properties, CFSTR(kIOBlockStorageDriverStatisticsKey));
 
-		if (statistics) {
+                    if (statistics) {
 			/* Obtain the number of bytes read from the drive statistics */
 			number = (CFNumberRef) CFDictionaryGetValue (statistics,
 								     CFSTR(kIOBlockStorageDriverStatisticsBytesReadKey));
@@ -2282,10 +2256,11 @@ getDISKcounters()
 				status = CFNumberGetValue(number, kCFNumberSInt64Type, &value);
 				totalWriteBytes += value;
 			}
-		}
-		/* Release resources */
+                    }
+                    /* Release resources */
 
-		CFRelease(properties); properties = 0;
+                    CFRelease(properties); properties = 0;
+                }
 		IOObjectRelease(drive); drive = 0;
 	}
 	IOIteratorReset(drivelist);
